const App = @import("./App.zig");

const Device = @import("./gfx/Device.zig");

pub const Queue = @import("./gfx/Queue.zig");

pub const color = @import("./gfx/color.zig");

const coral = @import("coral");

const ext = @import("./ext.zig");

const msg = @import("./msg.zig");

const std = @import("std");

pub const Display = struct {
	sdl_window: *ext.SDL_Window,
	clear_color: color.Value = color.black,
	device: Device,

	pub fn resize(self: Display, width: u16, height: u16) void {
		ext.SDL_SetWindowSize(self.sdl_window, width, height);
	}

	pub fn retitle(self: Display, title: []const u8) void {
		var sentineled_title = [_:0]u8{0} ** 255;

		@memcpy(sentineled_title[0 .. @min(title.len, sentineled_title.len)], title);
		ext.SDL_SetWindowTitle(self.sdl_window, &sentineled_title);
	}

	pub fn set_resizable(self: Display, resizable: bool) void {
		ext.SDL_SetWindowResizable(self.sdl_window, @intFromBool(resizable));
	}
};

pub const Error = error {
	SDLError,
};

pub const Handle = Queue.Handle;

pub const Input = union (enum) {
	key_up: Key,
	key_down: Key,

	pub const Key = enum (u32) {
		a = ext.SDL_SCANCODE_A,
		d = ext.SDL_SCANCODE_D,
		s = ext.SDL_SCANCODE_S,
		w = ext.SDL_SCANCODE_W,

		pub fn scancode(key: Key) u32 {
			return @intFromEnum(key);
		}
	};
};

pub const MeshPrimitives = struct {
	quad_mesh: Handle,
};

pub fn poll(app: coral.Write(App), inputs: msg.Send(Input)) !void {
	var event = @as(ext.SDL_Event, undefined);

	while (ext.SDL_PollEvent(&event) != 0) {
		switch (event.type) {
			ext.SDL_QUIT => app.res.quit(),
			ext.SDL_KEYUP => try inputs.push(.{.key_up = @enumFromInt(event.key.keysym.scancode)}),
			ext.SDL_KEYDOWN => try inputs.push(.{.key_down = @enumFromInt(event.key.keysym.scancode)}),
			else => {},
		}
	}
}

pub fn setup(world: *coral.World, events: App.Events) (Error || std.Thread.SpawnError || std.mem.Allocator.Error)!void {
	if (ext.SDL_Init(ext.SDL_INIT_VIDEO) != 0) {
		return error.SDLError;
	}

	const sdl_window = create: {
		const position = ext.SDL_WINDOWPOS_CENTERED;
		const flags = ext.SDL_WINDOW_OPENGL;
		const width = 640;
		const height = 480;

		break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
			return error.SDLError;
		};
	};

	errdefer ext.SDL_DestroyWindow(sdl_window);

	var device = try Device.init(sdl_window);

	errdefer device.deinit();

	try world.set_resource(.main, Display{
		.device = device,
		.sdl_window = sdl_window,
	});

	try world.on_event(events.pre_update, coral.system_fn(poll), .{.label = "poll gfx"});
	try world.on_event(events.exit, coral.system_fn(stop), .{.label = "stop gfx"});
	try world.on_event(events.finish, coral.system_fn(submit), .{.label = "submit gfx"});
}

pub fn stop(display: coral.WriteBlocking(Display)) void {
	display.res.device.deinit();
	ext.SDL_DestroyWindow(display.res.sdl_window);
}

pub fn submit(display: coral.WriteBlocking(Display)) void {
	display.res.device.submit(display.res.sdl_window, display.res.clear_color);
}
