const Queue = @import("./Queue.zig");

const color = @import("./color.zig");

const coral = @import("coral");

const draw_2d = @import("./shaders/draw_2d.glsl.zig");

const ext = @import("../ext.zig");

const sokol = @import("sokol");

const std = @import("std");

thread: std.Thread,
render_state: *RenderState,

const AtomicBool = std.atomic.Value(bool);

const RenderWork = struct {
	pipeline_2d: sokol.gfx.Pipeline,
	instance_2d_buffers: coral.stack.Sequential(sokol.gfx.Buffer),
	resources: coral.stack.Sequential(Resource),

	const Resource = union (enum) {
		mesh_2d: struct {
			index_count: u32,
			vertex_buffer: sokol.gfx.Buffer,
			index_buffer: sokol.gfx.Buffer,
		},
	};

	const buffer_indices = .{
		.mesh = 0,
		.instance = 1,
	};

	fn deinit(self: *RenderWork) void {
		sokol.gfx.destroyPipeline(self.pipeline_2d);

		for (self.instance_2d_buffers.values) |buffer| {
			sokol.gfx.destroyBuffer(buffer);
		}

		self.instance_2d_buffers.deinit();

		for (self.resources.values) |resource| {
			switch (resource) {
				.mesh_2d => |mesh_2d| {
					sokol.gfx.destroyBuffer(mesh_2d.vertex_buffer);
					sokol.gfx.destroyBuffer(mesh_2d.index_buffer);
				},
			}
		}

		self.resources.deinit();

		self.* = undefined;
	}

	fn init(allocator: std.mem.Allocator) RenderWork {
		return .{
			.pipeline_2d = sokol.gfx.makePipeline(.{
				.label = "2D drawing pipeline",

				.layout = .{
					.attrs = get: {
						var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

						attrs[draw_2d.ATTR_vs_mesh_xy] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.mesh,
						};

						attrs[draw_2d.ATTR_vs_instance_xbasis] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[draw_2d.ATTR_vs_instance_ybasis] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[draw_2d.ATTR_vs_instance_origin] = .{
							.format = .FLOAT2,
							.buffer_index = buffer_indices.instance,
						};

						attrs[draw_2d.ATTR_vs_instance_color] = .{
							.format = .UBYTE4N,
							.buffer_index = buffer_indices.instance,
						};

						attrs[draw_2d.ATTR_vs_instance_depth] = .{
							.format = .FLOAT,
							.buffer_index = buffer_indices.instance,
						};

						break: get attrs;
					},

					.buffers = get: {
						var buffers = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

						buffers[buffer_indices.instance].step_func = .PER_INSTANCE;

						break: get buffers;
					},
				},

				.shader = sokol.gfx.makeShader(draw_2d.draw2dShaderDesc(sokol.gfx.queryBackend())),
				.index_type = .UINT16,
			}),

			.instance_2d_buffers = .{.allocator = coral.heap.allocator},
			.resources = .{.allocator = allocator},
		};
	}

	fn process_draw_2d_commands(self: *RenderWork, commands: []const Queue.Buffer.Draw2DCommand, target: Queue.Target) std.mem.Allocator.Error!void {
		const max_instances = 512;
		var instance_2d_buffers_used = @as(usize, 0);

		sokol.gfx.applyPipeline(self.pipeline_2d);

		sokol.gfx.applyUniforms(.VS, draw_2d.SLOT_Screen, sokol.gfx.asRange(&draw_2d.Screen{
			.screen_size = .{target.width, target.height},
		}));

		for (commands) |command| {
			const mesh_2d = &self.resources.values[command.mesh_2d.index().?].mesh_2d;

			const instance_size = @sizeOf(Queue.Instance2D);
			const full_instance_buffer_count = command.instances.len / max_instances;

			for (0 .. full_instance_buffer_count) |i| {
				defer instance_2d_buffers_used += 1;

				if (instance_2d_buffers_used == self.instance_2d_buffers.len()) {
					const instance_2d_buffer = sokol.gfx.makeBuffer(.{
						.size = @sizeOf(Queue.Instance2D) * max_instances,
						.usage = .STREAM,
						.label = "2D drawing instance buffer",
					});

					errdefer sokol.gfx.destroyBuffer(instance_2d_buffer);

					try self.instance_2d_buffers.push(instance_2d_buffer);
				}

				sokol.gfx.applyBindings(.{
					.vertex_buffers = get_buffers: {
						var buffers = [_]sokol.gfx.Buffer{.{}} ** 8;

						buffers[buffer_indices.instance] = self.instance_2d_buffers.values[instance_2d_buffers_used];
						buffers[buffer_indices.mesh] = mesh_2d.vertex_buffer;

						break: get_buffers buffers;
					},

					.index_buffer = mesh_2d.index_buffer,
				});

				sokol.gfx.updateBuffer(self.instance_2d_buffers.values[instance_2d_buffers_used], .{
					.ptr = command.instances.ptr + (max_instances * i),
					.size = instance_size * max_instances,
				});

				sokol.gfx.draw(0, mesh_2d.index_count, max_instances);
			}

			defer instance_2d_buffers_used += 1;

			if (instance_2d_buffers_used == self.instance_2d_buffers.len()) {
				const instance_2d_buffer = sokol.gfx.makeBuffer(.{
					.size = @sizeOf(Queue.Instance2D) * max_instances,
					.usage = .STREAM,
					.label = "2D drawing instance buffer",
				});

				errdefer sokol.gfx.destroyBuffer(instance_2d_buffer);

				try self.instance_2d_buffers.push(instance_2d_buffer);
			}

			sokol.gfx.applyBindings(.{
				.vertex_buffers = get_buffers: {
					var buffers = [_]sokol.gfx.Buffer{.{}} ** 8;

					buffers[buffer_indices.instance] = self.instance_2d_buffers.values[instance_2d_buffers_used];
					buffers[buffer_indices.mesh] = mesh_2d.vertex_buffer;

					break: get_buffers buffers;
				},

				.index_buffer = mesh_2d.index_buffer,
			});

			const remaining_instances = command.instances.len % max_instances;

			sokol.gfx.updateBuffer(self.instance_2d_buffers.values[instance_2d_buffers_used], .{
				.ptr = command.instances.ptr + full_instance_buffer_count,
				.size = instance_size * remaining_instances,
			});

			sokol.gfx.draw(0, mesh_2d.index_count, @intCast(remaining_instances));
		}
	}

	fn process_queue(self: *RenderWork, buffer: *const Queue.Buffer, target: Queue.Target) std.mem.Allocator.Error!void {
		for (buffer.open_commands.values) |command| {
			switch (command.resource) {
				.texture => {

				},

				.mesh_2d => |mesh_2d| {
					const index_buffer = sokol.gfx.makeBuffer(.{
						.data = sokol.gfx.asRange(mesh_2d.indices),
						.type = .INDEXBUFFER,
					});

					const vertex_buffer = sokol.gfx.makeBuffer(.{
						.data = sokol.gfx.asRange(mesh_2d.vertices),
						.type = .VERTEXBUFFER,
					});

					errdefer {
						sokol.gfx.destroyBuffer(index_buffer);
						sokol.gfx.destroyBuffer(vertex_buffer);
					}

					if (mesh_2d.indices.len > std.math.maxInt(u32)) {
						return error.OutOfMemory;
					}

					try self.resources.push(.{
						.mesh_2d = .{
							.index_buffer = index_buffer,
							.vertex_buffer = vertex_buffer,
							.index_count = @intCast(mesh_2d.indices.len),
						},
					});
				},
			}
		}

		try self.process_draw_2d_commands(buffer.draw_2d_commands.values, target);
	}
};

const RenderState = struct {
	finished: std.Thread.Semaphore = .{},
	is_running: AtomicBool = AtomicBool.init(true),
	ready: std.Thread.Semaphore = .{},
	clear_color: color.Value = color.compress(color.black),
	pixel_width: c_int = 0,
	pixel_height: c_int = 0,
};

const Self = @This();

pub fn deinit(self: *Self) void {
	self.render_state.is_running.store(false, .monotonic);
	self.render_state.ready.post();
	self.thread.join();
	coral.heap.allocator.destroy(self.render_state);

	self.* = undefined;
}

pub fn init(sdl_window: *ext.SDL_Window) (std.mem.Allocator.Error || std.Thread.SpawnError)!Self {
	const render_state = try coral.heap.allocator.create(RenderState);

	errdefer coral.heap.allocator.destroy(render_state);

	render_state.* = .{};

	const self = Self{
		.thread = spawn_thread: {
			const thread = try std.Thread.spawn(.{}, run, .{sdl_window, render_state});

			thread.setName("Ona Graphics") catch {
				std.log.warn("failed to name the graphics thread", .{});
			};

			break: spawn_thread thread;
		},

		.render_state = render_state,
	};

	self.submit(sdl_window, .{0, 0, 0, 1});

	return self;
}

fn run(sdl_window: *ext.SDL_Window, render_state: *RenderState) !void {
	var result = @as(c_int, 0);

	result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_FLAGS, ext.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG);
	result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_PROFILE_MASK, ext.SDL_GL_CONTEXT_PROFILE_CORE);
	result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MAJOR_VERSION, 3);
	result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MINOR_VERSION, 3);
	result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_DOUBLEBUFFER, 1);

	if (result != 0) {
		std.log.err("failed to set necessary OpenGL flags in graphics", .{});
	}

	const context = ext.SDL_GL_CreateContext(sdl_window);

	defer ext.SDL_GL_DeleteContext(context);

	render_state.finished.post();

	sokol.gfx.setup(.{
		.environment = .{
			.defaults = .{
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.sample_count = 1,
			},
		},

		.logger = .{
			.func = sokol.log.func,
		},
	});

	defer sokol.gfx.shutdown();

	var render_work = RenderWork.init(coral.heap.allocator);

	defer render_work.deinit();

	while (render_state.is_running.load(.monotonic)) {
		render_state.ready.wait();

		defer render_state.finished.post();

		sokol.gfx.beginPass(.{
			.swapchain = .{
				.width = render_state.pixel_width,
				.height = render_state.pixel_height,
				.sample_count = 1,
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.gl = .{.framebuffer = 0},
			},

			.action = .{
				.colors = get: {
					var actions = [_]sokol.gfx.ColorAttachmentAction{.{}} ** 4;

					actions[0] = .{
						.load_action = .CLEAR,
						.clear_value = @as(sokol.gfx.Color, @bitCast(render_state.clear_color)),
					};

					break: get actions;
				},
			}
		});

		try Queue.consume_submitted(Queue.Consumer.bind(RenderWork, &render_work, RenderWork.process_queue), .{
			.width = @floatFromInt(render_state.pixel_width),
			.height = @floatFromInt(render_state.pixel_height),
		});

		sokol.gfx.endPass();
		sokol.gfx.commit();
		ext.SDL_GL_SwapWindow(sdl_window);
	}
}

pub fn submit(self: Self, sdl_window: *ext.SDL_Window, clear_color: color.Value) void {
	self.render_state.finished.wait();
	ext.SDL_GL_GetDrawableSize(sdl_window, &self.render_state.pixel_width, &self.render_state.pixel_height);
	std.debug.assert(self.render_state.pixel_width > 0 and self.render_state.pixel_height > 0);

	self.render_state.clear_color = clear_color;

	Queue.swap();
	self.render_state.ready.post();
}
