const color = @import("./color.zig");

const coral = @import("coral");

const std = @import("std");

buffer: *Buffer,
primitives: *const Primitives,

const AtomicCount = std.atomic.Value(usize);

pub const Buffer = struct {
	arena: std.heap.ArenaAllocator,
	closed_handles: coral.stack.Sequential(usize),
	open_commands: coral.stack.Sequential(OpenCommand),
	draw_2d_commands: coral.stack.Sequential(Draw2DCommand),
	close_commands: coral.stack.Sequential(CloseCommand),

	pub const CloseCommand = struct {
		handle: Handle,
	};

	pub const Draw2DCommand = struct {
		instances: []const Instance2D,
		mesh_2d: Handle,
	};

	pub const OpenCommand = struct {
		handle: Handle = .none,
		label: ?[]const u8 = null,

		resource: union (enum) {
			texture: Texture,
			mesh_2d: Mesh2D,
		},

		pub const Mesh2D = struct {
			vertices: []const Vertex2D,
			indices: []const u16,
		};

		pub const Texture = struct {
			width: u16,
			height: u16,
			format: Format,
			access: Access,

			pub const Access = enum {
				static,
			};

			pub const Format = enum {
				rgba8888,
				bgra8888,
				argb8888,
				rgb888,
				bgr888,

				pub fn byte_size(self: Format) usize {
					return switch (self) {
						.rgba8888, .bgra8888, .argb8888 => 4,
						.rgb888, .bgr888 => 3,
					};
				}
			};
		};
	};

	fn deinit(self: *Buffer) void {
		self.arena.deinit();
		self.closed_handles.deinit();
		self.open_commands.deinit();
		self.draw_2d_commands.deinit();
		self.close_commands.deinit();
	}

	fn init(allocator: std.mem.Allocator) Buffer {
		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
			.closed_handles = .{.allocator = allocator},
			.open_commands = .{.allocator = allocator},
			.draw_2d_commands = .{.allocator = allocator},
			.close_commands = .{.allocator = allocator},
		};
	}

	pub fn clear(self: *Buffer) void {
		self.close_commands.clear();
		self.draw_2d_commands.clear();
		self.open_commands.clear();

		if (!self.arena.reset(.retain_capacity)) {
			std.log.warn("failed to reset the buffer of a gfx queue with retained capacity", .{});
		}
	}

	pub fn draw_2d(self: *Buffer, command: Draw2DCommand) std.mem.Allocator.Error!void {
		try self.draw_2d_commands.push(.{
			.instances = try self.arena.allocator().dupe(Instance2D, command.instances),
			.mesh_2d = command.mesh_2d,
		});
	}

	pub fn open(self: *Buffer, command: OpenCommand) std.mem.Allocator.Error!Handle {
		const reserved_handle = @as(Handle, switch (command.handle) {
			.none => @enumFromInt(reserve_handle: {
				if (self.closed_handles.get()) |handle| {
					std.debug.assert(self.closed_handles.pop());

					break: reserve_handle handle;
				}

				break: reserve_handle next_handle.fetchAdd(1, .monotonic);
			}),

			_ => |handle| handle,
		});

		std.debug.assert(reserved_handle != .none);

		const arena_allocator = self.arena.allocator();

		try self.open_commands.push(.{
			.resource = switch (command.resource) {
				.texture => |texture| .{
					.texture = .{
						.width = texture.width,
						.height = texture.height,
						.format = texture.format,
						.access = texture.access,
					},
				},

				.mesh_2d => |mesh_2d| .{
					.mesh_2d = .{
						.indices = try arena_allocator.dupe(u16, mesh_2d.indices),
						.vertices = try arena_allocator.dupe(Vertex2D, mesh_2d.vertices),
					},
				},
			},

			.label = if (command.label) |label| try arena_allocator.dupe(coral.io.Byte, label) else null,
			.handle = reserved_handle,
		});

		return reserved_handle;
	}
};

pub const Consumer = coral.io.Generator(std.mem.Allocator.Error!void, &.{*const Buffer, Target});

pub const Handle = enum (usize) {
	none,
	_,

	pub fn index(self: Handle) ?usize {
		return switch (self) {
			.none => null,
			_ => @intFromEnum(self) - 1,
		};
	}
};

pub const Instance2D = extern struct {
	xbasis: Point2D = .{1, 0},
	ybasis: Point2D = .{0, 1},
	origin: Point2D = @splat(0),
	color: color.Compressed = color.compress(color.white),
	depth: f32 = 0,
};

const Node = struct {
	buffers: [2]Buffer,
	is_swapped: bool = false,
	ref_count: AtomicCount = AtomicCount.init(1),
	has_next: ?*Node = null,
	has_prev: ?*Node = null,

	fn acquire(self: *Node) void {
		self.ref_count.fetchAdd(1, .monotonic);
	}

	fn pending(self: *Node) *Buffer {
		return &self.buffers[@intFromBool(self.is_swapped)];
	}

	fn release(self: *Node) void {
		if (self.ref_count.fetchSub(1, .monotonic) == 1) {
			mutex.lock();

			defer mutex.unlock();

			if (self.has_prev) |prev| {
				prev.has_next = self.has_next;
			} else {
				has_head = self.has_next;
			}

			if (self.has_next) |next| {
				next.has_prev = self.has_prev;
			} else {
				has_tail = self.has_prev;
			}

			for (&self.buffers) |*buffer| {
				buffer.deinit();
			}

			coral.heap.allocator.destroy(self);
		}
	}

	fn submitted(self: *Node) *Buffer {
		return &self.buffers[@intFromBool(!self.is_swapped)];
	}
};

pub const Point2D = @Vector(2, f32);

pub const Primitives = struct {
	quad_mesh: Handle,
};

const Self = @This();

pub const State = struct {
	node: *Node,
	primitives: *const Primitives,
};

pub const Target = struct {
	width: f32,
	height: f32,
};

pub const Vertex2D = struct {
	xy: Point2D,
};

pub fn bind(context: coral.system.BindContext) std.mem.Allocator.Error!State {
	const queue = try coral.heap.allocator.create(Node);

	errdefer coral.heap.allocator.destroy(queue);

	queue.* = .{
		.buffers = .{Buffer.init(coral.heap.allocator), Buffer.init(coral.heap.allocator)},
	};

	mutex.lock();

	defer mutex.unlock();

	if (has_tail) |tail| {
		tail.has_next = queue;
		queue.has_prev = tail;
	} else {
		std.debug.assert(has_head == null);

		has_head = queue;
	}

	has_tail = queue;

	return .{
		.primitives = (try context.register_read_only_resource_access(.none, Primitives)) orelse create: {
			const buffer = queue.pending();
			const half_extent = 0.5;

			try context.world.set_resource(.none, Primitives{
				.quad_mesh = try buffer.open(.{
					.label = "quad mesh primitive",

					.resource = .{
						.mesh_2d = .{
							.indices = &.{0, 1, 2, 0, 2, 3},

							.vertices = &.{
								.{.xy = .{-half_extent, half_extent}},// .uv = .{0, 1}},
								.{.xy = .{half_extent, half_extent}},// .uv = .{1, 1}},
								.{.xy = .{half_extent, -half_extent}},// .uv = .{1, 0}},
								.{.xy = .{-half_extent, -half_extent}},//  .uv = .{0, 0}},
							},
						},
					},
				}),
			});

			break: create (try context.register_read_only_resource_access(.none, Primitives)).?;
		},

		.node = queue,
	};
}

pub fn consume_submitted(consumer: Consumer, target: Target) std.mem.Allocator.Error!void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;
	var iterations = @as(usize, 0);

	while (has_node) |node| : ({
		has_node = node.has_next;
		iterations += 1;
	}) {
		const buffer = &node.buffers[@intFromBool(!node.is_swapped)];

		try consumer.yield(.{buffer, target});

		buffer.clear();
	}
}

var has_head = @as(?*Node, null);

var has_tail = @as(?*Node, null);

pub fn init(state: *State) Self {
	return .{
		.buffer = state.node.pending(),
		.primitives = state.primitives,
	};
}

var mutex = std.Thread.Mutex{};

var next_handle = AtomicCount.init(1);

pub fn swap() void {
	mutex.lock();

	defer mutex.unlock();

	var has_node = has_head;

	while (has_node) |node| : (has_node = node.has_next) {
		node.is_swapped = !node.is_swapped;
	}
}

pub fn unbind(state: *State) void {
	state.node.release();
}
